@@ -17,21 +17,25 @@
 
 import collections
 import copy
+import functools
 import os
 import shutil
 import sys
+from typing import Any, Mapping, Optional, Sequence, Union
 
 from absl import flags
 from absl import logging
 from absl.testing import absltest
 import numpy as np
-from t5.data import sentencepiece_vocabulary
+from t5.data import dataset_providers
+from t5.data import preprocessors
 from t5.data import utils as dataset_utils
-import tensorflow.compat.v1 as tf
+from t5.data import vocabularies
+import tensorflow.compat.v2 as tf
 import tensorflow_datasets as tfds
 
-TaskRegistry = dataset_utils.TaskRegistry
-MixtureRegistry = dataset_utils.MixtureRegistry
+TaskRegistry = dataset_providers.TaskRegistry
+MixtureRegistry = dataset_providers.MixtureRegistry
 
 mock = absltest.mock
 
@@ -219,7 +223,12 @@ def _assert_compare_to_fake_dataset(
         {"id": [], "ids": [None], "idx": [], "idxs": [None]})
   _pyunit_proxy.assertDictEqual(
       expected_output_shapes,
-      {k: v.as_list() for k, v in ds.output_shapes.items()})
+      {k: v.shape.as_list() for k, v in ds.element_spec.items()})
+
+  expected_output_dtypes = {k: f.dtype for k, f in features.items()}
+  _pyunit_proxy.assertDictContainsSubset(
+      expected_output_dtypes,
+      {k: v.dtype for k, v in ds.element_spec.items()})
 
   actual_examples = _get_comparable_examples_from_ds(ds)
   expected_examples = [
@@ -229,12 +238,22 @@ def _assert_compare_to_fake_dataset(
 
 def verify_task_matches_fake_datasets(
     task, use_cached, token_preprocessed=False, splits=("train", "validation"),
+    num_shards=None
 ):
   """Assert all splits for both tokenized datasets are correct."""
   for split in splits:
+    get_dataset = functools.partial(
+        task.get_dataset, _SEQUENCE_LENGTH, split, use_cached=use_cached,
+        shuffle=False)
+    if num_shards:
+      ds = get_dataset(shard_info=dataset_providers.ShardInfo(0, num_shards))
+      for i in range(1, num_shards):
+        ds = ds.concatenate(
+            get_dataset(shard_info=dataset_providers.ShardInfo(i, num_shards)))
+    else:
+      ds = get_dataset()
     _assert_compare_to_fake_dataset(
-        task.get_dataset(
-            _SEQUENCE_LENGTH, split, use_cached=use_cached, shuffle=False),
+        ds,
         split,
         task.output_features,
         token_preprocessed=token_preprocessed,
@@ -262,12 +281,23 @@ def dataset_as_text(ds):
     yield {k: _maybe_as_text(v) for k, v in ex.items()}
 
 
-def assert_dataset(dataset, expected):
-  """Tests whether the entire dataset == expected or [expected]."""
+def assert_dataset(
+    dataset: tf.data.Dataset,
+    expected: Union[Mapping[str, Any], Sequence[Mapping[str, Any]]],
+    expected_dtypes: Optional[Mapping[str, tf.DType]] = None):
+  """Tests whether the entire dataset == expected or [expected].
+
+  Args:
+    dataset: a tf.data dataset
+    expected: either a single example, or a list of examples. Each example is a
+      dictionary.
+    expected_dtypes: an optional mapping from feature key to expected dtype.
+  """
+
   if not isinstance(expected, list):
     expected = [expected]
-  dataset = list(tfds.as_numpy(dataset))
-  _pyunit_proxy.assertEqual(len(dataset), len(expected))
+  actual = list(tfds.as_numpy(dataset))
+  _pyunit_proxy.assertEqual(len(actual), len(expected))
 
   def _compare_dict(actual_dict, expected_dict):
     _pyunit_proxy.assertEqual(
@@ -281,23 +311,62 @@ def assert_dataset(dataset, expected):
       np.testing.assert_array_equal(
           actual_value, _maybe_as_bytes(expected_dict[key]), key)
 
-  for actual_ex, expected_ex in zip(dataset, expected):
+  for actual_ex, expected_ex in zip(actual, expected):
     _compare_dict(actual_ex, expected_ex)
 
+  if expected_dtypes:
+    actual_dtypes = {k: dataset.element_spec[k].dtype for k in expected_dtypes}
+    _pyunit_proxy.assertDictEqual(expected_dtypes, actual_dtypes)
 
-def get_fake_dataset(split, shuffle_files=False):
+
+def assert_datasets_eq(dataset1: tf.data.Dataset, dataset2: tf.data.Dataset):
+  """Assert that two tfds datasets are equal."""
+
+  dataset1 = list(tfds.as_numpy(dataset1))
+  dataset2 = list(tfds.as_numpy(dataset2))
+  _pyunit_proxy.assertEqual(len(dataset1), len(dataset2))
+
+  def _compare_dict(dataset1, dataset2):
+    _pyunit_proxy.assertEqual(
+        set(dataset1.keys()), set(dataset2.keys()))
+    for key, value1 in dataset1.items():
+      if isinstance(value1, dict):
+        _compare_dict(value1, dataset2[key])
+        continue
+      if isinstance(value1, tf.RaggedTensor):
+        value1 = value1.to_list()
+      np.testing.assert_array_equal(
+          value1, _maybe_as_bytes(dataset2[key]), key)
+
+  for ex1, ex2 in zip(dataset1, dataset2):
+    _compare_dict(ex1, ex2)
+
+
+def assert_datasets_neq(dataset1, dataset2):
+  """Assert that two tfds datasets are unequal."""
+
+  _pyunit_proxy.assertRaises(AssertionError,
+                             assert_datasets_eq, dataset1, dataset2)
+
+
+def get_fake_dataset(split, shuffle_files=False, seed=None, shard_info=None):
   """Returns a tf.data.Dataset with fake data."""
   del shuffle_files  # Unused, to be compatible with TFDS API.
+  del seed
+
   output_types = {"prefix": tf.string, "suffix": tf.string}
   if split == "validation":
     output_types.update(
-        {"idx": tf.int64, "idxs": tf.int64, "id": tf.string, "ids": tf.string})
+        {"idx": tf.int32, "idxs": tf.int32, "id": tf.string, "ids": tf.string})
   output_shapes = {k: [] for k in output_types}
   if split == "validation":
     output_shapes.update({"idxs": [None], "ids": [None]})
 
-  return tf.data.Dataset.from_generator(
+  ds = tf.data.Dataset.from_generator(
       lambda: _FAKE_DATASET[split], output_types, output_shapes)
+  if shard_info:
+    ds = ds.shard(num_shards=shard_info.num_shards, index=shard_info.index)
+  return ds
 
 
 def test_text_preprocessor(dataset):
@@ -313,7 +382,7 @@ def test_text_preprocessor(dataset):
     })
     return res
 
-  return dataset.map(my_fn)
+  return dataset.map(my_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)
 
 
 def _split_tsv_preprocessor(dataset, field_names=("prefix", "suffix")):
@@ -327,36 +396,70 @@ def _split_tsv_preprocessor(dataset, field_names=("prefix", "suffix")):
             field_delim="\t", use_quote_delim=False)
     ))
 
-  return dataset.map(parse_line)
+  return dataset.map(
+      parse_line, num_parallel_calls=tf.data.experimental.AUTOTUNE)
 
 
-def test_token_preprocessor(dataset, output_features, **unused_kwargs):
+def test_token_preprocessor(dataset, output_features, sequence_length):
   """Change all occurrences of non-zero even numbered tokens in inputs to 50."""
   del output_features
+  del sequence_length
 
   def my_fn(ex):
     inputs = ex["inputs"]
     res = ex.copy()
-    res["inputs"] = tf.where_v2(
+    res["inputs"] = tf.where(
         tf.greater(inputs, 15),
-        tf.constant(50, tf.int64),
+        tf.constant(50, inputs.dtype),
         inputs)
     return res
 
-  return dataset.map(my_fn)
+  return dataset.map(my_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)
 
 
-def mock_vocabulary(encode_dict, vocab_size=None):
-  vocab = mock.MagicMock()
-  vocab.vocab_size = vocab_size
-  vocab.encode = mock.MagicMock(side_effect=lambda x: encode_dict[x])
-  vocab.encode_tf = mock.MagicMock(
-      side_effect=lambda x: tf.constant(encode_dict[x]))
-  return vocab
+@dataset_utils.map_over_dataset(num_seeds=1)
+def random_token_preprocessor(ex, seed):
+  """Selects a random shift to roll the tokens by for each feature."""
+  for feat in ["inputs", "targets"]:
+    tokens = ex[feat]
+    res = ex.copy()
+    n_tokens = tf.size(tokens)
+    random_shift = tf.random.stateless_uniform(
+        [], maxval=n_tokens, dtype=tf.int32, seed=seed)
+    res[feat] = tf.roll(tokens, shift=random_shift, axis=0)
+  return res
+
+
+def token_preprocessor_no_sequence_length(dataset, output_features):
+  return test_token_preprocessor(dataset, output_features, sequence_length=None)
+
+
+class MockVocabulary(object):
+  """Mocks a vocabulary object for testing."""
+
+  def __init__(self, encode_dict, vocab_size=None):
+    self._encode_dict = encode_dict
+    self._vocab_size = vocab_size
+
+  def encode(self, s):
+    return self._encode_dict[s]
+
+  def encode_tf(self, s):
+    res = tf.constant([-1], tf.int32)
+    for k, v in self._encode_dict.items():
+      if tf.equal(s, k):
+        res = tf.constant(v, tf.int32)
+      else:
+        pass
+    return res
+
+  @property
+  def vocab_size(self):
+    return self._vocab_size
 
 
 def sentencepiece_vocab(extra_ids=0):
-  return sentencepiece_vocabulary.SentencePieceVocabulary(
+  return vocabularies.SentencePieceVocabulary(
       os.path.join(TEST_DATA_DIR, "sentencepiece", "sentencepiece.model"),
       extra_ids=extra_ids)
 
@@ -369,12 +472,11 @@ def add_tfds_task(
     splits=None):
   TaskRegistry.add(
       name,
-      dataset_utils.TfdsTask,
+      dataset_providers.TfdsTask,
       tfds_name=tfds_name,
       text_preprocessor=text_preprocessor,
       token_preprocessor=token_preprocessor,
-      sentencepiece_model_path=os.path.join(TEST_DATA_DIR, "sentencepiece",
-                                            "sentencepiece.model"),
+      output_features=dataset_providers.Feature(sentencepiece_vocab()),
       metric_fns=[],
       splits=splits)
 
@@ -386,14 +488,14 @@ def add_task(
     token_preprocessor=None,
     splits=("train", "validation"),
     **kwargs):
+  if "output_features" not in kwargs:
+    kwargs["output_features"] = dataset_providers.Feature(sentencepiece_vocab())
   TaskRegistry.add(
       name,
       dataset_fn=dataset_fn,
       splits=splits,
       text_preprocessor=text_preprocessor,
       token_preprocessor=token_preprocessor,
-      sentencepiece_model_path=os.path.join(
-          TEST_DATA_DIR, "sentencepiece", "sentencepiece.model"),
       metric_fns=[],
       **kwargs)
 
@@ -454,7 +556,9 @@ class FakeTaskTest(absltest.TestCase):
                 "take": -1
             }],
     }
-    def _load_shard(shard_instruction):
+    def _load_shard(shard_instruction, shuffle_files, seed):
+      del shuffle_files
+      del seed
       fname = shard_instruction["filename"]
       if "train" in fname:
         if fname.endswith("00000-of-00002"):
@@ -486,9 +590,10 @@ class FakeTaskTest(absltest.TestCase):
     # Register a cached test Task.
     dataset_utils.set_global_cache_dirs([self.test_data_dir])
     clear_tasks()
-    add_tfds_task("cached_task")
+    add_tfds_task("cached_task", token_preprocessor=test_token_preprocessor)
+    add_tfds_task("cached_task_no_token_prep")
 
-    # Prepare cached task.
+    # Prepare cached tasks.
     self.cached_task = TaskRegistry.get("cached_task")
     cached_task_dir = os.path.join(self.test_data_dir, "cached_task")
     _dump_fake_dataset(
@@ -497,10 +602,18 @@ class FakeTaskTest(absltest.TestCase):
     _dump_fake_dataset(
         os.path.join(cached_task_dir, "validation.tfrecord"),
         _FAKE_TOKENIZED_DATASET["validation"], [2], _dump_examples_to_tfrecord)
+    shutil.copytree(
+        cached_task_dir,
+        os.path.join(self.test_data_dir, "cached_task_no_token_prep"))
 
     # Prepare uncached TfdsTask.
-    add_tfds_task("uncached_task")
+    add_tfds_task("uncached_task", token_preprocessor=test_token_preprocessor)
+    add_tfds_task("uncached_task_no_token_prep")
     self.uncached_task = TaskRegistry.get("uncached_task")
+    # Prepare uncached, random TfdsTask
+    add_tfds_task("uncached_random_task",
+                  token_preprocessor=random_token_preprocessor)
+    self.uncached_random_task = TaskRegistry.get("uncached_random_task")
 
     # Prepare uncached TextLineTask.
     _dump_fake_dataset(
@@ -508,17 +621,96 @@ class FakeTaskTest(absltest.TestCase):
         _FAKE_DATASET["train"], [2, 1], _dump_examples_to_tsv)
     TaskRegistry.add(
         "text_line_task",
-        dataset_utils.TextLineTask,
+        dataset_providers.TextLineTask,
         split_to_filepattern={
             "train": os.path.join(self.test_data_dir, "train.tsv*"),
         },
         skip_header_lines=1,
         text_preprocessor=[_split_tsv_preprocessor, test_text_preprocessor],
-        sentencepiece_model_path=os.path.join(
-            TEST_DATA_DIR, "sentencepiece", "sentencepiece.model"),
+        output_features=dataset_providers.Feature(sentencepiece_vocab()),
         metric_fns=[])
     self.text_line_task = TaskRegistry.get("text_line_task")
 
+    # Prepare uncached TFExampleTask
+    _dump_fake_dataset(
+        os.path.join(self.test_data_dir, "train.tfrecord"),
+        _FAKE_DATASET["train"], [2, 1], _dump_examples_to_tfrecord)
+    TaskRegistry.add(
+        "tf_example_task",
+        dataset_providers.TFExampleTask,
+        split_to_filepattern={
+            "train": os.path.join(self.test_data_dir, "train.tfrecord*"),
+        },
+        feature_description={
+            "prefix": tf.io.FixedLenFeature([], tf.string),
+            "suffix": tf.io.FixedLenFeature([], tf.string),
+        },
+        text_preprocessor=[test_text_preprocessor],
+        output_features=dataset_providers.Feature(sentencepiece_vocab()),
+        metric_fns=[])
+    self.tf_example_task = TaskRegistry.get("tf_example_task")
+
+    # Prepare uncached Task.
+    def _dataset_fn(split,
+                    shuffle_files,
+                    filepattern=os.path.join(self.test_data_dir, "train.tsv*")):
+      del split
+      files = tf.data.Dataset.list_files(filepattern, shuffle=shuffle_files)
+      return files.interleave(
+          lambda f: tf.data.TextLineDataset(f).skip(1),
+          num_parallel_calls=tf.data.experimental.AUTOTUNE)
+
+    TaskRegistry.add(
+        "general_task",
+        dataset_providers.Task,
+        dataset_fn=_dataset_fn,
+        splits=["train"],
+        text_preprocessor=[_split_tsv_preprocessor, test_text_preprocessor],
+        output_features=dataset_providers.Feature(sentencepiece_vocab()),
+        metric_fns=[])
+    self.general_task = TaskRegistry.get("general_task")
+
+    # Prepare uncached TaskV3.
+    TaskRegistry.add(
+        "task_v3",
+        dataset_providers.TaskV3,
+        source=dataset_providers.FunctionSource(
+            dataset_fn=get_fake_dataset,
+            splits=["train", "validation"]
+        ),
+        preprocessors=[
+            test_text_preprocessor,
+            preprocessors.tokenize,
+            token_preprocessor_no_sequence_length,
+            dataset_providers.CacheDatasetPlaceholder(),
+        ],
+        output_features={
+            "inputs": dataset_providers.Feature(sentencepiece_vocab()),
+            "targets": dataset_providers.Feature(sentencepiece_vocab()),
+        },
+        metric_fns=[])
+    self.task_v3 = TaskRegistry.get("task_v3")
+
+    # Prepare uncached TaskV3 with no caching before tokenization.
+    TaskRegistry.add(
+        "task_v3_tokenized_postcache",
+        dataset_providers.TaskV3,
+        source=dataset_providers.FunctionSource(
+            dataset_fn=get_fake_dataset,
+            splits=["train", "validation"]
+        ),
+        preprocessors=[
+            test_text_preprocessor,
+            dataset_providers.CacheDatasetPlaceholder(),
+            preprocessors.tokenize,
+            token_preprocessor_no_sequence_length,
+        ],
+        output_features={
+            "inputs": dataset_providers.Feature(sentencepiece_vocab()),
+            "targets": dataset_providers.Feature(sentencepiece_vocab()),
+        },
+        metric_fns=[])
+
   def tearDown(self):
     super().tearDown()
     self._tfds_patcher.stop()
@@ -541,3 +733,9 @@ class FakeMixtureTest(FakeTaskTest):
         [("cached_task", 1.0)],
     )
     self.cached_mixture = MixtureRegistry.get("cached_mixture")
+    MixtureRegistry.add(
+        "uncached_random_mixture",
+        [("uncached_random_task", 1.0)],
+    )
+    self.uncached_mixture = MixtureRegistry.get(
+        "uncached_random_mixture")
